/*
    $Id: getmntent.c,v 1.2 1994/05/31 23:56:57 marion Exp $
    getmntent.c - replacement for (set end get)^mntent.
    May 94 by marion@cs.tu-berlin.de

    Copyright (C) 1994 Klaus Rennecke

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* $Log: getmntent.c,v $
 * Revision 1.2  1994/05/31  23:56:57  marion
 * Put getnfswd and companions into libgetnfswd.a.
 * Added --version option to nfspwd.
 * Added check to set $MAKE variable.
 * Check for getwd system call.
 * More ANSIfying done.
 * Supply path name for mount program in getmntent.c.
 * Ported to sinix and DEC alpha. sinix is a special nuisance.
 *
 * Revision 1.1  1994/05/30  16:55:06  marion
 * Put under autoconf control.
 * Ported to ULTRIX 4.[23].
 * Added getmntent emulation for ULTRIX.
 * ANSIfied the sources.
 *
 */

#include <stdio.h>

#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#include <sys/param.h>

#include "getmntent.h"

/*
 * This is defined only for specific system types, since the output of
 * mount(8) is quite system dependend. It might be quite easy to port, though.
 */
#if defined(__ultrix) || defined(__alpha)

#if !defined(MOUNT_PROG)
#if defined(__ultrix)
#define MOUNT_PROG	"/bin/mount"
#else
#if defined(__alpha)
#define MOUNT_PROG	"/usr/sbin/mount"
#endif
#endif
#endif

FILE *
setmntent ()
{
    return popen (MOUNT_PROG, "r");	/* call mount(8) */
}


struct mntent *
#if defined(__STDC__)
getmntent(FILE *fp)
#else /* __STDC__ */
getmntent(fp)
FILE	*fp;
#endif /* __STDC__ */
{
    static mntent ent;			/* beware - static return value */
    static char inbuf[MAXPATHLEN*2+40];	/* Two pathnames, type and options */
    char *fill;				/* to parse filler words in output */
    
    if (!fgets (inbuf, sizeof (inbuf), fp))
    	return 0;
    
    /*
     * Parse mount(8) output:
     *
     * /dev/rz3a on / type ufs
     * vogon:/export/dec on /tmp_amd/vogon/export/dec type nfs (rw,hard,ping=6000,timeo=10,retrans=5,intr,nosuid)
     *	...
     */
     
    if (!(ent.mnt_fsname = strtok (inbuf, " \t")))
    	return 0;
    if (!(fill = strtok ((char *)0, " \t")) || strcmp (fill, "on"))
    	return 0;
    if (!(ent.mnt_dir = strtok ((char *)0, " \t")))
    	return 0;
    if (!(fill = strtok ((char *)0, " \t")) || strcmp (fill, "type"))
    	return 0;
    if (!(ent.mnt_type = strtok ((char *)0, " \t\n")))
    	return 0;
	
    ent.mnt_opts = strtok ((char *)0, "\n"); /* optional */
    
    return &ent;
}


int
#if defined(__STDC__)
endmntent (FILE *fp)
#else /* __STDC__ */
endmntent (fp)
FILE	*fp;
#endif /* __STDC__ */
{
    return pclose (fp);
}


#endif /* __ultrix */
