/*
    $Id: getnfswd.c,v 1.11 1994/05/31 23:57:00 marion Exp $
    getnfswd.c - derive the nfs soft link path from unix getwd and
		 mount table entries from automounter.
    Feb 93 by marion@cs.tu-berlin.de

    Copyright (C) 1993, 1994 Klaus Rennecke

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* $Log: getnfswd.c,v $
 * Revision 1.11  1994/05/31  23:57:00  marion
 * Put getnfswd and companions into libgetnfswd.a.
 * Added --version option to nfspwd.
 * Added check to set $MAKE variable.
 * Check for getwd system call.
 * More ANSIfying done.
 * Supply path name for mount program in getmntent.c.
 * Ported to sinix and DEC alpha. sinix is a special nuisance.
 *
 * Revision 1.10  1994/05/30  18:36:42  marion
 * Added more header file dependent configuration.
 * Tar make target added.
 *
 * Revision 1.9  1994/05/30  16:55:13  marion
 * Put under autoconf control.
 * Ported to ULTRIX 4.[23].
 * Added getmntent emulation for ULTRIX.
 * ANSIfied the sources.
 *
 * Revision 1.8  1994/02/06  11:30:02  marion
 * Sigh - forgot ';'
 *
 * Revision 1.7  1994/02/06  11:27:18  marion
 * Added support for the BSD amd subdir option.
 *
 * Revision 1.6  1994/01/10  15:14:42  marion
 * Port to solaris.
 *
 * Revision 1.5  1993/06/29  08:44:05  marion
 * Fixed ugly memory leak in findlink.
 *
 * Revision 1.4  1993/06/29  07:59:21  marion
 * New mechanism for auto.home mounted filesystems - now robust against
 *   file system soft link names differing from last source path name
 *   component. This is a bit slower and does a lot of stat'ing, but it
 *   is far more correct.
 *
 * Revision 1.3  1993/06/29  06:07:28  marion
 * Put under GGPL.
 *
 * Revision 1.2  1993/02/21  19:07:18  marion
 * changed getnfswd to use /etc/mtab. This is a quite nice algorithm.
 * *pose*
 *
 */

#include <stdio.h>

#if defined(sinix) && !defined(__GNUC__)
/* The sinix c compiler uses a crude include path, which breaks the
   header file tests of configure. I was greatly tempted to neglect
   portability to this system, so expect this to vanish any time. */
#include <sys/mnttab.h>
#include <sys/mntent.h>
#define MOUNTED MNTTAB
#define setmntent(n,m)  fopen((n),(m))
#define endmntent(n)    fclose((n))
#define mnt_fsname      mnt_special
#define mnt_dir         mnt_mountp
#define mntent          mnttab
#else
#if HAVE_MNTENT_H && !HAVE_SYS_MNTTAB_H
#include <mntent.h>
#else /* HAVE_MNTENT_H */
#if defined(HAVE_SYS_MNTTAB_H)
#include <sys/mnttab.h>
/* #include <sys/mntent.h> probably obsolete solaris quirk */
#define MOUNTED MNTTAB
#define setmntent(n,m)	fopen((n),(m))
#else /* SYS_MNTTAB */
#include "getmntent.h"
#endif /* SYS_MNTTAB */
#endif /* HAVE_MNTENT_H */
#endif /* sinix destaster */

#include <netdb.h>
#include <unistd.h>

#include <malloc.h>
#include <string.h>

#include <sys/param.h>
#include <sys/types.h>

#include <sys/stat.h>
#include "scandir.h"

extern int gethostname();

static char hostname[MAXHOSTNAMELEN] = ""; /* local host name */

static char **automounted = 0;		/* automounter directory cache */
static char **mountpaths = 0;		/* mount point list cache */

static char *rcsid =
	"@(#) $Id: getnfswd.c,v 1.11 1994/05/31 23:57:00 marion Exp $";

    /* read_mtab - Read in /etc/mtab. Store the information in 'automounted'
		   'mountpaths' and the modified time in 'mtabtime'. On next
		   call, if /etc/mtab was not changed, we needn't read in the
		   table again.

	Input:	'name' is usually the local host name, to filter entries for
		locally mounted or automounter entries.

	Output:	0 if the table could not be read or memory allocation failed,
		else 1.
    */

static int
#if defined(__STDC__)
read_mtab (const char *name)
#else /* __STDC__ */
read_mtab (name)
char *name;
#endif /* __STDC__ */
{
    FILE	*mtab;			/* open mnttab */
#ifdef HAVE_SYS_MNTTAB_H

    /* This is *&^@$@ annoying !!! stop this stupidity of sys5 */
#define mntent mnttab
#define mnt_type mnt_fstype
#define mnt_dir mnt_mountp
#define mnt_fsname mnt_special
#define mnt_opts mnt_mntopts

    struct mntent mntbuf;		/* buffer for getmntent information */
#endif
    struct mntent *mnt;			/* getmntent information */
    int		nauto;			/* counter for automounted */
    int		nmounted;		/* counter for mountpaths */
    
#if !defined(MNTENT_NO_STAT)
    static time_t mtabmtime = 0;
    struct stat	st;			/* for mtime check */

    /* check if /etc/mtab changed */
    if (stat (MOUNTED, &st)) {
    	return 0;			/* failed stat */
    } else if (mtabmtime == st.st_mtime) {
    	return 1;			/* mtab not modified */
    }
    
    mtabmtime = st.st_mtime;		/* remember mtime */
    
    /* open /etc/mtab */
    if (!(mtab = setmntent (MOUNTED, "r"))) {
	return 0;
    }
#else /* MNTENT_NO_STAT */
    /* open /etc/mtab */
    if (!(mtab = setmntent ())) {
	return 0;
    }
#endif /* MNTENT_NO_STAT */
    
    /* free old entries in automounted */
    if (automounted) {
    	char **p;
	
	for (p = automounted; *p; p++)
	    free (*p);
	free (automounted);
	automounted = 0;
    }
    
    /* free old entries in mountpaths */
    if (mountpaths) {
    	char **p;
	
	for (p = mountpaths; *p; p++)
	    free (*p);
	free (mountpaths);
	mountpaths = 0;
    }
    
    /* search mtab for entries */
    nauto = 0;
    nmounted = 0;
#ifndef HAVE_SYS_MNTTAB_H
    while ((mnt = getmntent (mtab)))
#else
    mnt = &mntbuf;
    while (!getmntent (mtab, mnt))
#endif
    {
    	char *p;
	
	if ((p = strchr (mnt->mnt_fsname, ':'))) {
	    if (!strncmp (mnt->mnt_fsname, name, p - mnt->mnt_fsname)) {
	    	if (!automounted) {
		    automounted = (char **)malloc (2 * sizeof (char**));
		} else {
		    automounted = (char **)realloc (automounted,
			(nauto+2) * sizeof (char**));
		}
		automounted[nauto] = malloc (strlen (mnt->mnt_dir)+1);
		strcpy (automounted[nauto], mnt->mnt_dir);
		nauto++;
		automounted[nauto] = 0;
	    } else {
	    	if (!mountpaths) {
		    mountpaths = (char **)malloc (2 * sizeof (char**));
		} else {
		    mountpaths = (char **)realloc (mountpaths,
			(nmounted+2) * sizeof (char**));
		}
		mountpaths[nmounted] = malloc (strlen (mnt->mnt_dir)+1);
		strcpy (mountpaths[nmounted], mnt->mnt_dir);
		nmounted++;
		mountpaths[nmounted] = 0;
	    }
	}
    }
    
#if !defined(__svr4__) || !defined(sun) /* !solaris */
    endmntent (mtab);
#endif
    return 1;
}

/* findlink - find a link in prefix `stat'ing to match, prefix is storage */
static char *
#if defined(__STDC__)
findlink (char * prefix, struct stat * match)
#else /* __STDC__ */
findlink (prefix, match)
char		*prefix;
struct stat	*match;
#endif /* __STDC__ */
{
    static struct dirent **namelist = 0;/* 1-entry directory cache */
    static int		n_namelist;	/* entries in directory */
    static struct stat	oldhome;	/* old stat result of /home */
    struct stat		sthome;		/* stat result in /home */
    char		*suffix;	/* suffix part pointer */
    int			i;		/* loop counter */

    suffix = prefix + strlen (prefix);

    /* check if exists */
    if (stat (prefix, &sthome)) {
	/* does not exist */
	return 0;
    }

    /* check if it matches the cache entry */
    if (!namelist ||
	sthome.st_dev != oldhome.st_dev ||
	sthome.st_ino != oldhome.st_ino ||
	sthome.st_mtime != oldhome.st_mtime) {

	/* mismatch - re-read directory */
	if (namelist) {
	    for (i = 0; i < n_namelist; i++) {
		if (namelist[i]) {
		    free (namelist[i]);
		}
	    }
	    free (namelist);
	    namelist = 0;
	}

	n_namelist = scandir (prefix, &namelist, (int (*)())0, (int (*)())0);

	if (n_namelist < 3) {
	    /* only . and .. are not enough */
	    return 0;
	}

	oldhome = sthome;
    }

    *suffix = '/';

    /* loop through the entries */
    for (i = 0; i < n_namelist; i++) {

	/* check if exists */
	strcpy (suffix+1, namelist[i]->d_name);

	if (stat (prefix, &sthome)) {
	    /* does not exist */
	    continue;
	}

	/* compare the stat results */
	if (sthome.st_dev == match->st_dev &&
	    sthome.st_ino == match->st_ino) {

	    /* prefix now contains the full path name of the link */
	    return prefix;
	}
    }

    /* not found - recover the prefix */
    *suffix = 0;
    return 0;
}

    /* getnfswd - A replacement for getwd(3) which tries to recover the
		  original path name where the automounter linked the file
		  system.

	Output:	The current directory, either as the result of getwd, or
		edited for automouned file systems if the link in the
		automounter directory is found and matched.
    */

char *
#if defined(__STDC__)
getnfswd (char path[MAXPATHLEN])
#else /* __STDC__ */
getnfswd (path)
char path[MAXPATHLEN];
#endif /* __STDC__ */
{
#if HAVE_GETWD
    extern char *getwd ();		/* the original */
#else
    extern char *getcwd ();		/* the original */
#define getwd(p)	getcwd(p,MAXPATHLEN)
#endif
    char	home[MAXPATHLEN];	/* path name in local directory */
    struct stat	sthome;			/* stat result in local directory */
    struct stat	stnfs;			/* stat result nfs mounted */
    char	**homes;		/* scan through automounter list */
    char	*ret;			/* return value */

    if ((ret = getwd (path))) {
	/* get local hostname */
    	if (!hostname[0] && gethostname (hostname, sizeof (hostname))) {
	    return 0;
	}
	
	/* read in entries from /etc/mtab, if modified */
	if (!read_mtab (hostname)) {
	    return 0;
	}
	
	/* loop through automounter directories */
	for (homes = automounted; homes && *homes; homes++) {
	    char *cont;
	    char **mount;		/* scan through mount points */

	    /* copy automounter path */
	    strcpy (home, *homes);
	    cont = home + strlen (home); /* end of home path */
	    
	    /* loop through nfs mount points */
	    for (mount = mountpaths; mount && *mount; mount++) {
		char *link;
		int len;
		
		/* is path mounted at mount ? */
		len = strlen (*mount);
		if (strncmp (*mount, path, len) ||
		    (path[len] != '/' && path[len] != 0)) {
		    continue;
		}

		/* check for existence of mount point */
		if (stat (*mount, &stnfs)) {
		    continue;
		}

		if ((link = strchr (*mount+1, '/')) &&
		    !strncmp (link, home, strlen (home))) {
		    if (!stat (link, &sthome) &&
			sthome.st_dev == stnfs.st_dev &&
			sthome.st_ino == stnfs.st_ino) {
			/* append the rest to home */
			strcat (home, path+(link-*mount)+strlen(home));
			/* copy to return buffer */
			strcpy (path, home);
			return ret;
		    }
		}

		/* find possible link name in auto.home mounted fashion */
		if (findlink (home, &stnfs)) {
		    /* from auto.home: append name to mount link */
		    strcat (home, path+len);
		    /* copy to return buffer */
		    strcpy (path, home);
		    return ret;
		}

		/* next possibility: from auto.direct */

		/* check if exists */
		if (stat (home, &sthome)) {
		    /* does not exist */
    	    	    continue;
		}

		/* compare the stat results */
		if (sthome.st_dev == stnfs.st_dev &&
		    sthome.st_ino == stnfs.st_ino) {
		    /* append the rest to home */
		    strcat (home, path+len);
		    /* copy to return buffer */
		    strcpy (path, home);
		    return ret;
		}

		/* next possibility: amd subdir option */
		if (path[len] == '/' && path[len+1] &&
		    (link = strchr (path+len+1, '/'))) {
		    /* append next path component and /.. to home */
		    strncpy (cont, path+len, (link-path)-len);
		    cont[(link-path)-len] = 0;
		    strcat (cont, "/..");

		    /* check if link and subdir exists */
		    if (!stat (home, &sthome) &&
			sthome.st_dev == stnfs.st_dev &&
			sthome.st_ino == stnfs.st_ino) {

			/* append the rest to home */
			strcpy (cont+(link-path)-len, link);
			/* copy to return buffer */
			strcpy (path, home);
			    return ret;

		    } else {
			/* repair home path */
			*cont = 0;
		    }
		}
		
		/*
		 * We may think it would be wise to break here, because
		 * we already know that the path was mounted at mount and
		 * is unlikely to be mounted elsewhere. But you _may_ mount
		 * a nfs file system below the mount point of another.
		 */
	    }
	}
    }
	
    return ret;
    
    (void)*rcsid;
}
