/*
    $Id: scandir.c,v 1.4 1994/05/31 23:57:05 marion Exp $
    scandir.c - sys5 substitute for scandir.
    Feb 93 by marion@cs.tu-berlin.de

    Copyright (C) 1993, 1994 Klaus Rennecke

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* $Log: scandir.c,v $
 * Revision 1.4  1994/05/31  23:57:05  marion
 * Put getnfswd and companions into libgetnfswd.a.
 * Added --version option to nfspwd.
 * Added check to set $MAKE variable.
 * Check for getwd system call.
 * More ANSIfying done.
 * Supply path name for mount program in getmntent.c.
 * Ported to sinix and DEC alpha. sinix is a special nuisance.
 *
 * Revision 1.3  1994/05/30  18:36:46  marion
 * Added more header file dependent configuration.
 * Tar make target added.
 *
 * Revision 1.2  1994/05/30  16:55:21  marion
 * Put under autoconf control.
 * Ported to ULTRIX 4.[23].
 * Added getmntent emulation for ULTRIX.
 * ANSIfied the sources.
 *
 * Revision 1.1  1994/01/10  15:14:44  marion
 * Port to solaris.
 *
 */

#if STDC_HEADERS
#include <stdlib.h>
#endif

#if STDC_HEADERS || HAVE_STRING_H
#include <string.h>
/* An ANSI string.h and pre-ANSI memory.h might conflict.  */
#if !STDC_HEADERS && HAVE_MEMORY_H
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
/* memory.h and strings.h conflict on some systems.  */
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#include "scandir.h"

int
#if defined(__STDC__)
scandir (const char *path, struct dirent ***namelist,
	 int (*select)(), int (*compare)())
#else /* __STDC__ */
scandir (path, namelist, select, compare)
char		*path;
struct dirent	***namelist;
int		(*select)();
int		(*compare)();
#endif /* __STDC__ */
{
    struct dirent	*entry;
    struct dirent	**newlist;
    int			n_entry = 0;
    DIR			*dir;

    if (!(dir = opendir (path))) {
	return -1;
    }

    if (!(newlist = (struct dirent **)malloc (sizeof (*newlist)))) {
	return -1;
    }

    while ((entry = readdir (dir))) {
	if (!select || (*select)(entry)) {
	    n_entry++;
	    newlist = (struct dirent **)
	    	realloc (newlist, (n_entry + 1) * sizeof (*newlist));
	    newlist[n_entry-1] = entry;
	}
    }
    newlist[n_entry] = 0;

    if (compare) {
	qsort (newlist, (size_t)n_entry, sizeof (*newlist), compare);
    }

    *namelist = newlist;
    return n_entry;
}
